# coding=utf-8

import logging

from xpaw.utils.project import load_object

log = logging.getLogger(__name__)


class MiddlewareManager:
    def __init__(self, *middlewares):
        self._open_handlers = []
        self._close_handlers = []
        for middleware in middlewares:
            self._add_middleware(middleware)

    @classmethod
    def _middleware_list_from_config(cls, config):
        raise NotImplementedError

    @classmethod
    def from_config(cls, config):
        mw_list = cls._middleware_list_from_config(config)
        mws = []
        for cls_path in mw_list:
            mw_cls = load_object(cls_path)
            if hasattr(mw_cls, "from_config"):
                mw = mw_cls.from_config(config)
            else:
                mw = mw_cls()
            mws.append(mw)
        return cls(*mws)

    def _add_middleware(self, middleware):
        if hasattr(middleware, "open"):
            self._open_handlers.append(middleware.open)
        if hasattr(middleware, "close"):
            self._close_handlers.append(middleware.close)

    def open(self):
        for method in self._open_handlers:
            method()

    def close(self):
        for method in self._close_handlers:
            method()



# coding=utf-8


class NetworkError(Exception):
    """
    Error occurred when download.
    """


class IgnoreRequest(Exception):
    """
    Ignore this request.
    """


class ResponseNotMatch(Exception):
    """
    The response is not as desired.
    """


class UsageError(Exception):
    """
    CLI usage error.
    """

    def __init__(self, *args, **kwargs):
        self.print_help = kwargs.pop('print_help', True)
        super().__init__(*args, **kwargs)



# coding=utf-8

import logging
import asyncio

import aiohttp
import async_timeout

from xpaw.middleware import MiddlewareManager
from xpaw.http import HttpRequest, HttpResponse
from xpaw.errors import NetworkError

log = logging.getLogger(__name__)


class Downloader:
    def __init__(self, timeout=None, loop=None):
        self._timeout = timeout
        self._loop = loop or asyncio.get_event_loop()

    async def download(self, request):
        log.debug("HTTP request: {} {}".format(request.method, request.url))
        timeout = request.meta.get("timeout")
        if not timeout:
            timeout = self._timeout
        async with aiohttp.ClientSession(cookies=request.cookies,
                                         loop=self._loop,
                                         cookie_jar=request.meta.get("cookie_jar")) as session:
            with async_timeout.timeout(timeout, loop=self._loop):
                async with session.request(request.method,
                                           request.url,
                                           headers=request.headers,
                                           data=request.body,
                                           proxy=request.proxy) as resp:
                    body = await resp.read()
                    cookies = resp.cookies
        response = HttpResponse(resp.url,
                                resp.status,
                                headers=resp.headers,
                                body=body,
                                cookies=cookies)
        log.debug("HTTP response: {} {}".format(response.url, response.status))
        return response


class DownloaderMiddlewareManager(MiddlewareManager):
    def __init__(self, *middlewares):
        self._request_handlers = []
        self._response_handlers = []
        self._error_handlers = []
        super().__init__(*middlewares)

    def _add_middleware(self, middleware):
        super()._add_middleware(middleware)
        if hasattr(middleware, "handle_request"):
            self._request_handlers.append(self._coro_wrapper(middleware.handle_request))
        if hasattr(middleware, "handle_response"):
            self._response_handlers.append(self._coro_wrapper(middleware.handle_response))
        if hasattr(middleware, "handle_error"):
            self._error_handlers.append(self._coro_wrapper(middleware.handle_error))

    def _coro_wrapper(self, func):
        if asyncio.iscoroutinefunction(func):
            return func

        async def coro_wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if asyncio.iscoroutine(result):
                result = await result
            return result

        return coro_wrapper

    @classmethod
    def _middleware_list_from_config(cls, config):
        mw_list = config.get("downloader_middlewares")
        if mw_list:
            if not isinstance(mw_list, list):
                mw_list = [mw_list]
        else:
            mw_list = []
        log.debug("Downloader middleware list: {}".format(mw_list))
        return mw_list

    async def download(self, downloader, request):
        try:
            res = await self._handle_request(request)
            if isinstance(res, HttpRequest):
                return res
            if res is None:
                try:
                    response = await downloader.download(request)
                except Exception as e:
                    log.debug("Network error {}: {}".format(type(e), e))
                    raise NetworkError(e)
                else:
                    res = response
            _res = await self._handle_response(request, res)
            if _res:
                res = _res
        except Exception as e:
            try:
                res = await self._handle_error(request, e)
            except Exception:
                raise
            else:
                if res is not True:
                    if res:
                        return res
                    raise e
        else:
            return res

    async def _handle_request(self, request):
        for method in self._request_handlers:
            res = await method(request)
            if not (res is None or isinstance(res, (HttpRequest, HttpResponse))):
                raise TypeError("Request handler must return None, HttpRequest or HttpResponse,"
                                " got {}".format(type(res)))
            if res:
                return res

    async def _handle_response(self, request, response):
        for method in self._response_handlers:
            res = await method(request, response)
            if not (res is None or isinstance(res, HttpRequest)):
                raise TypeError("Response handler must return None or HttpRequest,"
                                " got {}".format(type(res)))
            if res:
                return res

    async def _handle_error(self, request, error):
        handled = False
        for method in self._error_handlers:
            res = await method(request, error)
            if not (res is None or res is True or isinstance(res, (HttpRequest, HttpResponse))):
                raise TypeError("Exception handler must return None, True, HttpRequest or HttpResponse,"
                                " got {}".format(type(res)))
            if res is True:
                handled = True
            elif res:
                return res
        return handled



# coding=utf-8

import pytest
import json

import asyncio

from xpaw import HttpRequest
from xpaw.downloader import Downloader, DownloaderMiddlewareManager
from xpaw.downloadermws import CookieJarMiddleware


@pytest.fixture(scope="module")
def loop(request):
    def close():
        loop.close()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    request.addfinalizer(close)
    return loop


@pytest.fixture(scope="module")
def downloader(loop):
    return Downloader(timeout=20, loop=loop)


def test_cookies(downloader, loop):
    req = HttpRequest("http://httpbin.org/cookies", cookies={"k1": "v1", "k2": "v2"})
    resp = loop.run_until_complete(downloader.download(req))
    cookies = json.loads(resp.text)["cookies"]
    assert len(cookies) == 2 and cookies.get("k1") == "v1" and cookies.get("k2") == "v2"


def test_cookie_jar(downloader, loop):
    dmm = DownloaderMiddlewareManager(CookieJarMiddleware(loop=loop))
    loop.run_until_complete(dmm.download(downloader, HttpRequest("http://httpbin.org/cookies/set?k1=v1&k2=v2")))
    resp = loop.run_until_complete(dmm.download(downloader, HttpRequest("http://httpbin.org/cookies")))
    cookies = json.loads(resp.text)["cookies"]
    assert len(cookies) == 2 and cookies.get("k1") == "v1" and cookies.get("k2") == "v2"
    loop.run_until_complete(dmm.download(downloader, HttpRequest("http://httpbin.org/cookies/delete?k1=")))
    resp = loop.run_until_complete(dmm.download(downloader, HttpRequest("http://httpbin.org/cookies")))
    cookies = json.loads(resp.text)["cookies"]
    assert len(cookies) == 1 and cookies.get("k2") == "v2"



# coding=utf-8

import logging
import asyncio

from aiohttp import CookieJar

log = logging.getLogger(__name__)


class CookieJarMiddleware:
    def __init__(self, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._cookie_jar = CookieJar(loop=self._loop)

    @classmethod
    def from_config(cls, config):
        return cls(loop=config.get("downloader_loop"))

    async def handle_request(self, request):
        log.debug("Assign cookie jar to request (url={})".format(request.url))
        request.meta["cookie_jar"] = self._cookie_jar



# coding=utf-8

import yaml


def load_config_file(file):
    with open(file, "r", encoding="utf-8") as f:
        d = yaml.load(f)
        return d


def dump_config_file(file, config):
    with open(file, "w", encoding="utf-8") as f:
        yaml.dump(config, f)



# coding=utf-8

from xpaw.downloadermws.cookies import CookieJarMiddleware
from xpaw.downloadermws.forward import ForwardedForMiddleware
from xpaw.downloadermws.headers import RequestHeadersMiddleware
from xpaw.downloadermws.proxy import ProxyMiddleware, ProxyAgentMiddleware
from xpaw.downloadermws.retry import RetryMiddleware, ResponseMatchMiddleware
from xpaw.downloadermws.speed import SpeedLimitMiddleware



# coding=utf-8

import logging

log = logging.getLogger(__name__)


class RequestHeadersMiddleware:
    def __init__(self, headers):
        self._headers = headers or {}

    @classmethod
    def from_config(cls, config):
        return cls(config.get("request_headers"))

    async def handle_request(self, request):
        log.debug("Assign headers to request (url={}): {}".format(request.url, self._headers))
        for i in self._headers:
            request.headers[i] = self._headers[i]



# coding=utf-8

import logging
import asyncio

from xpaw.config import BaseConfig

log = logging.getLogger(__name__)


class SpeedLimitMiddleware:
    def __init__(self, span, burst, loop=None):
        if span <= 0:
            raise ValueError("span must greater than 0")
        self._span = span
        self._burst = burst
        self._loop = loop or asyncio.get_event_loop()
        self._semaphore = asyncio.Semaphore(0, loop=loop)
        self._value = 0
        self._update_future = None

    @classmethod
    def from_config(cls, config):
        c = config.get("speed_limit")
        if c is None:
            c = {}
        c = BaseConfig(c)
        return cls(c.getfloat("span", 1),
                   c.getint("burst", 1),
                   loop=config.get("downloader_loop"))

    async def handle_request(self, request):
        await self._semaphore.acquire()
        self._value -= 1

    async def _update_value(self):
        while True:
            log.debug("Update speed limit semaphore")
            d = self._burst - self._value
            if d > 0:
                self._value += d
                i = 0
                while i < d:
                    self._semaphore.release()
                    i += 1
            await asyncio.sleep(self._span)

    def open(self):
        self._update_future = asyncio.ensure_future(self._update_value(), loop=self._loop)

    def close(self):
        if self._update_future:
            self._update_future.cancel()



# coding=utf-8


import xpaw
from xpaw.commands import Command


class VersionCommand(Command):
    @property
    def name(self):
        return "version"

    @property
    def short_desc(self):
        return "Print the version"

    def add_arguments(self, parser):
        Command.add_arguments(self, parser)

    def run(self, args):
        print("xpaw version {}".format(xpaw.__version__))



# coding=utf-8

import os
import sys
import logging

from bson import ObjectId

from xpaw.config import Config
from xpaw.utils.project import load_object
from xpaw.downloader import DownloaderMiddlewareManager
from xpaw.spider import SpiderMiddlewareManager
from xpaw.utils.config import load_config_file

log = logging.getLogger(__name__)


class TaskLoader:
    def __init__(self, proj_dir, base_config=None, **kwargs):
        # add project path
        sys.path.append(proj_dir)
        # copy sys.modules
        modules_keys = set(sys.modules.keys())
        self.config = self._load_task_config(proj_dir, base_config)
        for k, v in kwargs.items():
            self.config.set(k, v, "project")
        # task id
        if "task_id" not in self.config:
            self.config.set("task_id", "{}".format(ObjectId()), "project")
        self.downloadermw = DownloaderMiddlewareManager.from_config(self.config)
        self.spider = load_object(self.config["spider"])(self.config)
        self.spidermw = SpiderMiddlewareManager.from_config(self.config)
        # recover sys.modules
        keys = list(sys.modules.keys())
        for k in keys:
            if k not in modules_keys:
                del sys.modules[k]
        # remove project path
        sys.path.remove(proj_dir)

    def _load_task_config(self, proj_dir, base_config=None):
        c = load_config_file(os.path.join(proj_dir, "config.yaml"))
        log.debug("Project specific configuration: {}".format(c))
        task_config = base_config or Config()
        task_config.update(c, priority="project")
        return task_config

    def open_spider(self):
        self.spidermw.open()
        self.downloadermw.open()

    def close_spider(self):
        self.downloadermw.close()
        self.spidermw.close()



# coding=utf-8

import re
import cgi


def get_encoding_from_header(content_type):
    content_type, params = cgi.parse_header(content_type)
    if "charset" in params:
        return params["charset"]


_charset_re = re.compile(r"""<meta.*?charset=["']*(.+?)["'>]""", flags=re.I)
_pragma_re = re.compile(r"""<meta.*?content=["']*;?charset=(.+?)["'>]""", flags=re.I)
_xml_re = re.compile(r"""^<\?xml.*?encoding=["']*(.+?)["'>]""")


def get_encoding_from_content(content):
    if isinstance(content, bytes):
        content = content.decode("ascii", errors="ignore")
    elif not isinstance(content, str):
        raise ValueError("content should be bytes or str")
    s = _charset_re.search(content)
    if s:
        return s.group(1).strip()
    s = _pragma_re.search(content)
    if s:
        return s.group(1).strip()
    s = _xml_re.search(content)
    if s:
        return s.group(1).strip()



# coding=utf-8

from os.path import join, isfile, abspath
import logging.config

from xpaw.errors import UsageError
from xpaw.commands import Command
from xpaw.cluster import LocalCluster
from xpaw.utils.log import configure_logging

log = logging.getLogger(__name__)


class CrawlCommand(Command):
    @property
    def syntax(self):
        return "<project_dir>"

    @property
    def name(self):
        return "crawl"

    @property
    def short_desc(self):
        return "Start to crawl web pages"

    def add_arguments(self, parser):
        Command.add_arguments(self, parser)

        parser.add_argument("project_dir", metavar="project_dir", nargs="?", help="project directory")

    def process_arguments(self, args):
        Command.process_arguments(self, args)

    def run(self, args):
        if not args.project_dir:
            raise UsageError()
        if not isfile(join(args.project_dir, "config.yaml")):
            self.exitcode = 1
            print("Error: Connot find 'config.yaml' in {}".format(abspath(args.project_dir)))
            return
        
        configure_logging("xpaw", self.config)
        cluster = LocalCluster(args.project_dir, self.config)
        cluster.start()



# coding=utf-8

import os
from os.path import exists, join, abspath, isdir, basename
from shutil import move, copy2, copystat, ignore_patterns

from xpaw.commands import Command
from xpaw.errors import UsageError
from xpaw.utils.template import string_camelcase, render_templatefile

IGNORE = ignore_patterns("*.pyc")


class InitCommand(Command):
    def __init__(self):
        super().__init__()

        self.steps_total = 0
        self.steps_count = 0

    @property
    def syntax(self):
        return "<project_dir>"

    @property
    def name(self):
        return "init"

    @property
    def short_desc(self):
        return "Initialize a crawling project"

    def add_arguments(self, parser):
        Command.add_arguments(self, parser)

        parser.add_argument("project_dir", metavar="project_dir", nargs="?", help="project directory")

    def process_arguments(self, args):
        Command.process_arguments(self, args)

        if args.project_dir:
            if not exists(args.project_dir):
                os.mkdir(args.project_dir, mode=0o775)

    def run(self, args):
        if not args.project_dir:
            raise UsageError()

        project_dir = abspath(args.project_dir)
        project_name = basename(project_dir)

        if exists(join(project_dir, "config.yaml")):
            self.exitcode = 1
            print("Error: config.yaml already exists in {}".format(project_dir))
            return

        self._init_project(args.project_dir, project_name)

    def _init_project(self, project_dir, project_name):
        self._copytree(join(self.config["templates_dir"], "project"), project_dir)
        move(join(project_dir, "module"), join(project_dir, project_name))
        self._render_files(project_dir,
                           lambda f: render_templatefile(f, project_name=project_name,
                                                         ProjectName=string_camelcase(project_name)))

        print("New project '{}', created in:".format(project_name,
                                                     self.config["templates_dir"]))
        print("    {}".format(abspath(project_dir)))

    def _copytree(self, src, dst):
        if not exists(dst):
            os.makedirs(dst)
        copystat(src, dst)
        names = os.listdir(src)
        ignored_names = IGNORE(src, names)
        for name in names:
            if name in ignored_names:
                continue
            srcname = join(src, name)
            dstname = join(dst, name)
            if isdir(srcname):
                self._copytree(srcname, dstname)
            else:
                copy2(srcname, dstname)

    def _render_files(self, src, render):
        names = os.listdir(src)
        for name in names:
            srcname = join(src, name)
            if isdir(srcname):
                self._render_files(srcname, render)
            else:
                render(srcname)



# coding=utf-8

from os.path import abspath, join, dirname

LOG_FILE = None
LOG_ENCODING = "utf-8"
LOG_LEVEL = "INFO"
LOG_FORMAT = "%(asctime)s %(name)s [%(levelname)s]: %(message)s"
LOG_DATEFORMAT = "%Y-%m-%d %H:%M:%S"

TEMPLATES_DIR = abspath(join(dirname(__file__), "..", "templates"))

DOWNLOADER_CLIENTS = 100
DOWNLOADER_TIMEOUT = 20

QUEUE_CLS = "xpaw.queue.RequestDequeue"



# coding=utf-8

import re
import asyncio
import threading

import pytest
from aiohttp import web
from yarl import URL

from xpaw.config import Config
from xpaw.http import HttpRequest, HttpResponse
from xpaw.downloadermws import *
from xpaw.errors import IgnoreRequest, ResponseNotMatch

from .helpers import wait_server_start


@pytest.fixture(scope="module")
def loop(request):
    def close():
        loop.close()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    request.addfinalizer(close)
    return loop


class TestForwardedForMiddleware:
    def test_handle_request(self, loop):
        mw = ForwardedForMiddleware()
        req = HttpRequest("http://www.example.com")
        loop.run_until_complete(mw.handle_request(req))
        assert re.search(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", req.headers["X-Forwarded-For"])


class TestRequestHeadersMiddleware:
    def test_handle_request(self, loop):
        headers = {"Content-Type": "text/html", "User-Agent": "xpaw", "Connection": "keep-alive"}
        mw = RequestHeadersMiddleware.from_config(dict(request_headers=headers))
        req = HttpRequest("http://www.example.com")
        loop.run_until_complete(mw.handle_request(req))
        assert headers == req.headers


@pytest.fixture(scope="module")
def agent(request):
    async def handle_request(request):
        return web.Response(body=b'["127.0.0.1:3128", "127.0.0.1:8080"]')

    def handle_error(loop, context):
        pass

    def start_loop():
        app = web.Application(loop=loop)
        app.router.add_resource("/").add_route("GET", handle_request)
        loop.run_until_complete(loop.create_server(app.make_handler(access_log=None, loop=loop), "0.0.0.0", 7340))
        try:
            loop.run_forever()
        except Exception:
            pass
        finally:
            loop.close()

    def stop_loop():
        loop.call_soon_threadsafe(loop.stop)

    loop = asyncio.new_event_loop()
    loop.set_exception_handler(handle_error)
    t = threading.Thread(target=start_loop)
    t.start()
    wait_server_start("127.0.0.1:7340")
    request.addfinalizer(stop_loop)


class TestProxyMiddleware:
    def test_hanle_request(self, loop):
        mw = ProxyMiddleware(["127.0.0.1"])
        req = HttpRequest("http://www.example.com")
        loop.run_until_complete(mw.handle_request(req))
        assert req.proxy == "http://127.0.0.1"


class TestProxyAgentMiddleware:
    def test_handle_request(self, loop, monkeypatch):
        async def _pick_proxy():
            return "http://127.0.0.1"

        mw = ProxyAgentMiddleware.from_config(Config({"proxy_agent": {"addr": "http://127.0.0.1:7340"}}))
        monkeypatch.setattr(mw, "_pick_proxy", _pick_proxy)
        req = HttpRequest("http://www.example.com")
        loop.run_until_complete(mw.handle_request(req))
        assert req.proxy == "http://127.0.0.1"

    def test_pick_proxy(self, loop, monkeypatch):
        async def _update_proxy_list():
            while self.index < 2:
                await asyncio.sleep(0.1)
                self.index += 1
                mw._proxy_list = proxy_list[self.index - 1]

        self.index = 0
        proxy_list = [[], ["127.0.0.1", "127.0.0.2"]]
        res = ["http://127.0.0.1", "http://127.0.0.2"]
        mw = ProxyAgentMiddleware.from_config(
            Config({"proxy_agent": {"addr": "127.0.0.1:7340", "update_interval": 0.1}}))
        monkeypatch.setattr(mw, "_update_proxy_list", _update_proxy_list)
        mw.open()
        for i in range(len(res)):
            req = HttpRequest("http://www.example.com")
            loop.run_until_complete(mw.handle_request(req))
            assert req.proxy in res

    def test_update_proxy_list(self, loop, agent):
        async def _func():
            while mw._proxy_list is None:
                await asyncio.sleep(0.1)

        mw = ProxyAgentMiddleware.from_config(
            Config({"proxy_agent": {"addr": "http://127.0.0.1:7340", "update_interval": 0.1}}))
        mw.open()
        loop.run_until_complete(_func())
        assert mw._proxy_list == ["127.0.0.1:3128", "127.0.0.1:8080"]
        mw.close()
        loop.run_until_complete(asyncio.sleep(0.1))


class TestRetryMiddleware:
    def test_handle_reponse(self, loop, monkeypatch):
        class ErrorFlag(Exception):
            pass

        def _retry(request, reason):
            assert isinstance(request, HttpRequest) and isinstance(reason, str)
            raise ErrorFlag

        mw = RetryMiddleware.from_config(Config())
        monkeypatch.setattr(mw, "retry", _retry)
        req = HttpRequest("http://www.example.com")
        resp = HttpResponse(URL("http://www.example.com"), 400)
        loop.run_until_complete(mw.handle_response(req, resp))
        with pytest.raises(ErrorFlag):
            resp = HttpResponse(URL("http://www.example.com"), 503)
            loop.run_until_complete(mw.handle_response(req, resp))

    def test_handle_error(self, loop, monkeypatch):
        class ErrorFlag(Exception):
            pass

        def _retry(request, reason):
            assert isinstance(request, HttpRequest) and isinstance(reason, str)
            raise ErrorFlag

        mw = RetryMiddleware.from_config(Config())
        monkeypatch.setattr(mw, "retry", _retry)
        req = HttpRequest("http://www.example.com")
        err = ValueError()
        loop.run_until_complete(mw.handle_error(req, err))
        with pytest.raises(ErrorFlag):
            err = ResponseNotMatch()
            loop.run_until_complete(mw.handle_error(req, err))

    def test_retry(self):
        max_retry_times = 2
        mw = RetryMiddleware.from_config(Config({"retry": {"max_retry_times": max_retry_times}}))
        req = HttpRequest("http://www.example.com")
        for i in range(max_retry_times):
            req = mw.retry(req, "")
            assert isinstance(req, HttpRequest)
        with pytest.raises(IgnoreRequest):
            mw.retry(req, "")

    def test_match_status(self):
        assert RetryMiddleware.match_status("200", 200) is True
        assert RetryMiddleware.match_status(200, 200) is True
        assert RetryMiddleware.match_status("2xX", 201) is True
        assert RetryMiddleware.match_status("40x", 403) is True
        assert RetryMiddleware.match_status("40X", 403) is True
        assert RetryMiddleware.match_status("50x", 403) is False
        assert RetryMiddleware.match_status("~20X", 200) is False
        assert RetryMiddleware.match_status("!20x", 400) is True


class TestResponseMatchMiddleware:
    def test_handle_response(self, loop):
        req = HttpRequest("http://www.baidu.com")
        resp = HttpResponse(URL("http://www.baidu.com"), 200, body="<title>百度一下，你就知道</title>".encode("utf-8"))
        wrong_resp = HttpResponse(URL("http://www.qq.com"), 200, body="<title>腾讯QQ</title>".encode("utf-8"))
        mw = ResponseMatchMiddleware.from_config(Config({"response_match": {"url_pattern": "www\\.baidu\\.com",
                                                                            "body_pattern": "百度",
                                                                            "encoding": "utf-8"}}))
        loop.run_until_complete(mw.handle_response(req, resp))
        with pytest.raises(ResponseNotMatch):
            loop.run_until_complete(mw.handle_response(req, wrong_resp))


class TestCookieJarMiddleware:
    def test_handle_request(self, loop):
        mw = CookieJarMiddleware.from_config(Config({"downloader_loop": loop}))
        req = HttpRequest("http://www.example.com")
        loop.run_until_complete(mw.handle_request(req))
        assert req.meta.get("cookie_jar") is mw._cookie_jar



# coding=utf-8

import random
import logging

log = logging.getLogger(__name__)


class ForwardedForMiddleware:
    async def handle_request(self, request):
        x = "61.%s.%s.%s" % (random.randint(128, 191), random.randint(0, 255), random.randint(1, 254))
        log.debug("Assign 'X-Forwarded-For: {}' to request (url={})".format(x, request.url))
        request.headers["X-Forwarded-For"] = x



# coding=utf-8

from importlib import import_module
from pkgutil import iter_modules


def load_object(path):
    dot = path.rindex(".")
    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)
    return getattr(mod, name)


def walk_modules(path):
    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, "__path__"):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + "." + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods



# coding=utf-8

from lxml import etree


class Selector:
    def __init__(self, html=None, root=None):
        if html is not None:
            if not isinstance(html, str):
                raise TypeError("'html' argument should be of type {}".format(str))
            root = etree.HTML(html)
        elif root is None:
            raise ValueError("Selector needs either 'html' or 'root' argument")
        self._root = root

    def xpath(self, xpath):
        try:
            res = self._root.xpath(xpath, smart_strings=False)
        except Exception:
            return SelectorList([])
        if not isinstance(res, list):
            res = [res]
        return SelectorList([self.__class__(root=i) for i in res])

    @property
    def html(self):
        try:
            return etree.tostring(self._root, encoding="unicode", method="html", with_tail=False)
        except TypeError:
            return str(self._root)

    @property
    def text(self):
        try:
            return etree.tostring(self._root, encoding="unicode", method="text", with_tail=False)
        except TypeError:
            return str(self._root)


class SelectorList(list):
    def __getitem__(self, item):
        obj = super().__getitem__(item)
        return self.__class__(obj) if isinstance(item, slice) else obj

    def xpath(self, xpath):
        res = self.__class__()
        for i in self:
            res += i.xpath(xpath)
        return res

    @property
    def html(self):
        return [i.html for i in self]

    @property
    def text(self):
        return [i.text for i in self]



# coding=utf-8

import inspect

from xpaw.utils.web import get_encoding_from_header, get_encoding_from_content


class HttpRequest:
    def __init__(self, url, method="GET", body=None,
                 headers=None, cookies=None, proxy=None,
                 meta=None, callback=None):
        """
        Construct an HTTP request.
        """
        self.url = url
        self.method = method
        self.body = body
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.proxy = proxy
        self._meta = dict(meta) if meta else {}
        if callback and inspect.ismethod(callback):
            callback = callback.__name__
        self.callback = callback

    @property
    def meta(self):
        return self._meta

    def copy(self):
        kw = {}
        for x in ["url", "method", "body", "headers", "cookies", "proxy", "meta", "callback"]:
            kw.setdefault(x, getattr(self, x))
        return HttpRequest(**kw)


class HttpResponse:
    def __init__(self, url, status,
                 body=None, headers=None, cookies=None,
                 request=None):
        """
        Construct an HTTP response.
        """
        self.url = url
        self.status = status
        self.body = body
        self.headers = headers
        self.cookies = cookies
        self.request = request

    @property
    def encoding(self):
        if hasattr(self, "_encoding"):
            return self._encoding
        encoding = get_encoding_from_header(self.headers.get("content-type"))
        if not encoding and self.body:
            encoding = get_encoding_from_content(self.body)
        self._encoding = encoding or "utf-8"
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def text(self):
        if hasattr(self, "_text") and self._text:
            return self._text
        if not self.body:
            return ""
        encoding = self.encoding
        content = ""
        try:
            content = self.body.decode(encoding, errors="replace")
        except LookupError:
            content = self.body.decode(encoding, errors="replace")
        self._text = content
        return self._text

    @property
    def meta(self):
        if self.request:
            return self.request.meta
        return None

    def copy(self):
        kw = {}
        for x in ["url", "status", "body", "headers", "cookies", "request"]:
            kw.setdefault(x, getattr(self, x))
        return HttpResponse(**kw)



# coding=utf-8

import os
import re
import string


def render_templatefile(path, **kwargs):
    if path.endswith(".tmpl"):
        with open(path, "rb") as f:
            raw = f.read().decode("utf-8")

        content = string.Template(raw).substitute(**kwargs)

        render_path = path[:-len(".tmpl")]
        with open(render_path, "wb") as f:
            f.write(content.encode("utf-8"))
        os.remove(path)


CAMELCASE_INVALID_CHARS = re.compile('[^a-zA-Z\d]')


def string_camelcase(s):
    return CAMELCASE_INVALID_CHARS.sub('', s.title())



# coding=utf-8

from collections import deque


class RequestQueue:
    def push(self, request):
        raise NotImplementedError

    def pop(self):
        raise NotImplementedError


class RequestDequeue(RequestQueue):
    def __init__(self):
        self._queue = deque()

    def push(self, request):
        self._queue.append(request)

    def pop(self):
        if len(self._queue) > 0:
            return self._queue.popleft()



# coding=utf-8

from xpaw.errors import UsageError
from xpaw.config import Config


class Command:
    def __init__(self):
        self.config = Config()
        self.exitcode = 0

    @property
    def name(self):
        return ""

    @property
    def syntax(self):
        return ""

    @property
    def short_desc(self):
        return ""

    @property
    def long_desc(self):
        return self.short_desc

    def add_arguments(self, parser):
        parser.add_argument("-s", "--set", dest="set", action="append", default=[], metavar="NAME=VALUE",
                            help="set/override setting (may be repeated)")
        parser.add_argument("-l", "--log-level", dest="log_level", metavar="LEVEL",
                            help="log level")

    def process_arguments(self, args):
        # setting
        try:
            self.config.update(dict(x.split("=", 1) for x in args.set), priority="cmdline")
        except ValueError:
            raise UsageError("Invalid -s value, use -s NAME=VALUE", print_help=False)

        # logger
        if args.log_level:
            self.config.set("log_level", args.log_level, priority="cmdline")

    def run(self, args):
        raise NotImplementedError



# coding=utf-8

__all__ = ["__version__", "Spider", "HttpRequest", "HttpResponse", "Selector"]

from .version import __version__

# Add patch to avoid 'TIME_WAIT'
from . import _patch

del _patch

from xpaw.spider import Spider
from xpaw.http import HttpRequest, HttpResponse
from xpaw.selector import Selector



# coding=utf-8

import logging

from xpaw.http import HttpRequest

log = logging.getLogger(__name__)


class MaxDepthMiddleware:
    def __init__(self, max_depth):
        self._max_depth = max_depth

    @classmethod
    def from_config(cls, config):
        return cls(config.getint("max_depth", 0))

    def handle_output(self, response, result):
        depth = response.meta.get("_current_depth", 0) + 1
        for r in result:
            if isinstance(r, HttpRequest):
                if depth <= self._max_depth:
                    r.meta["_current_depth"] = depth
                    yield r
                else:
                    log.debug("The request(url={}) will be aborted as the depth of it is out of limit".format(r.url))
            else:
                yield r



# coding=utf-8

import time
import asyncio
import logging

from xpaw.downloader import Downloader
from xpaw.http import HttpRequest, HttpResponse
from xpaw.loader import TaskLoader
from xpaw.utils.project import load_object

log = logging.getLogger(__name__)


class LocalCluster:
    def __init__(self, proj_dir, config):
        self._config = config
        self._queue = load_object(self._config.get("queue_cls"))()
        self._downloader_loop = asyncio.new_event_loop()
        self._downloader_loop.set_exception_handler(self._handle_coro_error)
        self._downloader = Downloader(timeout=self._config.getfloat("downloader_timeout"),
                                      loop=self._downloader_loop)
        self._task_loader = TaskLoader(proj_dir, base_config=self._config, downloader_loop=self._downloader_loop)
        self._last_request = None
        self._futures = None

    def start(self):
        log.info("Task ID: {}".format(self._task_loader.config.get("task_id")))
        self._task_loader.open_spider()
        self._start_downloader_loop()

    def _handle_coro_error(self, loop, context):
        log.error("Unexpected error occurred when run the event loop: {}".format(context["message"]))

    async def _push_start_requests(self):
        for res in self._task_loader.spidermw.start_requests(self._task_loader.spider):
            if isinstance(res, HttpRequest):
                self._queue.push(res)
            elif isinstance(res, Exception):
                log.warning("Unexpected error occurred when handle start requests", exc_info=True)
            await asyncio.sleep(0.01, loop=self._downloader_loop)

    def _start_downloader_loop(self):
        self._futures = []
        f = asyncio.ensure_future(self._push_start_requests(), loop=self._downloader_loop)
        self._futures.append(f)
        asyncio.ensure_future(self._supervisor(), loop=self._downloader_loop)
        for i in range(self._task_loader.config.getint("downloader_clients")):
            f = asyncio.ensure_future(self._pull_requests(i), loop=self._downloader_loop)
            self._futures.append(f)

        asyncio.set_event_loop(self._downloader_loop)
        try:
            self._downloader_loop.run_forever()
        except Exception:
            log.error("Unexpected error occurred when run loop", exc_info=True)
            raise
        finally:
            log.info("Close event loop")
            self._downloader_loop.close()

    async def _supervisor(self):
        timeout = self._task_loader.config.getfloat("downloader_timeout")
        task_finished_delay = 2 * timeout
        self._last_request = time.time()
        while True:
            await asyncio.sleep(5, loop=self._downloader_loop)
            if time.time() - self._last_request > task_finished_delay:
                break
        try:
            self._task_loader.close_spider()
        except Exception:
            log.warning("Unexpected error occurred when close spider", exc_info=True)
        if self._futures:
            for f in self._futures:
                f.cancel()
            self._futures = None
        log.info("Event loop will be stopped after 3 seconds")
        await asyncio.sleep(3, loop=self._downloader_loop)
        self._downloader_loop.stop()

    async def _pull_requests(self, coro_id):
        while True:
            req = self._queue.pop()
            if req:
                self._last_request = time.time()
                log.debug("The request (url={}) has been pulled by coro[{}]".format(req.url, coro_id))
                try:
                    result = await self._task_loader.downloadermw.download(self._downloader, req)
                except Exception:
                    log.warning("Unexpected error occurred when request '{}'".format(req.url), exc_info=True)
                else:
                    self._handle_result(req, result)
            else:
                await asyncio.sleep(3, loop=self._downloader_loop)

    def _handle_result(self, request, result):
        if isinstance(result, HttpRequest):
            self._queue.push(result)
        elif isinstance(result, HttpResponse):
            # bind HttpRequest
            result.request = request
            try:
                for res in self._task_loader.spidermw.parse(self._task_loader.spider,
                                                            result):
                    if isinstance(res, HttpRequest):
                        self._queue.push(res)
            except Exception:
                log.warning("Unexpected error occurred when parse response of '{}'".format(request.url), exc_info=True)



# coding=utf-8

import logging


def configure_logging(name, config):
    logger = logging.getLogger(name)
    logger.setLevel(config["log_level"])
    log_stream_handler = logging.StreamHandler()
    log_stream_handler.setLevel(config["log_level"])
    log_formatter = logging.Formatter(config["log_format"], config["log_dateformat"])
    log_stream_handler.setFormatter(log_formatter)
    logger.addHandler(log_stream_handler)



# coding=utf-8

import json
import random
import asyncio
import logging

import aiohttp
import async_timeout

from xpaw.config import BaseConfig

log = logging.getLogger(__name__)


class ProxyMiddleware:
    def __init__(self, proxies):
        if not proxies or len(proxies) <= 0:
            raise ValueError("proxies cannot be empty")
        self._proxies = proxies
        self._n = len(proxies)

    @classmethod
    def from_config(cls, config):
        return cls(config.getlist("proxies"))

    async def handle_request(self, request):
        proxy = self._pick_proxy()
        log.debug("Assign proxy '{}' to request (url={})".format(proxy, request.url))
        if not proxy.startswith("http://"):
            proxy = "http://{}".format(proxy)
        request.proxy = proxy

    def _pick_proxy(self):
        i = random.randint(0, self._n - 1)
        return self._proxies[i]


class ProxyAgentMiddleware:
    def __init__(self, agent_addr, update_interval, update_timeout, loop=None):
        if not agent_addr.startswith("http://"):
            agent_addr = "http://{}".format(agent_addr)
        self._agent_addr = agent_addr
        self._update_interval = update_interval
        self._update_timeout = update_timeout
        self._loop = loop or asyncio.get_event_loop()
        self._proxy_list = None
        self._update_lock = asyncio.Lock(loop=self._loop)
        self._update_future = None

    @classmethod
    def from_config(cls, config):
        c = config.get("proxy_agent")
        if c is None:
            c = {}
        c = BaseConfig(c)
        return cls(c.get("addr"),
                   c.getfloat("update_interval", 60),
                   c.getfloat("update_timeout", 20),
                   loop=config.get("downloader_loop"))

    async def handle_request(self, request):
        proxy = await self._pick_proxy()
        log.debug("Assign proxy '{}' to request (url={})".format(proxy, request.url))
        if not proxy.startswith("http://"):
            proxy = "http://{}".format(proxy)
        request.proxy = proxy

    async def _pick_proxy(self):
        while True:
            async with self._update_lock:
                if self._proxy_list and len(self._proxy_list) > 0:
                    break
            await asyncio.sleep(self._update_interval, loop=self._loop)
        n = len(self._proxy_list)
        i = random.randint(0, n - 1)
        return self._proxy_list[i]

    async def _update_proxy_list(self):
        while True:
            log.debug("Update proxy list")
            try:
                async with self._update_lock:
                    async with aiohttp.ClientSession(loop=self._loop) as session:
                        with async_timeout.timeout(self._update_timeout, loop=self._loop):
                            async with session.get(self._agent_addr) as resp:
                                body = await resp.read()
                                proxy_list = json.loads(body.decode(encoding="utf-8"))
                                if proxy_list:
                                    self._proxy_list = proxy_list
            except Exception:
                log.warning("Unexpected error occurred when updating proxy list", exc_info=True)
            await asyncio.sleep(self._update_interval, loop=self._loop)

    def open(self):
        self._update_future = asyncio.ensure_future(self._update_proxy_list(), loop=self._loop)

    def close(self):
        if self._update_future:
            self._update_future.cancel()



# coding=utf-8

import copy
from collections import MutableMapping

from . import defaultconfig

CONFIG_PRIORITIES = {
    "default": 0,
    "module": 10,
    "project": 20,
    "cmdline": 30
}


def get_config_priority(priority):
    if isinstance(priority, str):
        return CONFIG_PRIORITIES[priority]
    else:
        return priority


class ConfigAttribute:
    def __init__(self, value, priority):
        self.value = value
        self.priority = priority

    def set(self, value, priority):
        if priority >= self.priority:
            if isinstance(self.value, BaseConfig):
                value = BaseConfig(value, priority=priority)
            self.value = value
            self.priority = priority

    def __str__(self):
        return "<ConfigAttribute value={self.value!r} priority={self.priority}>".format(self=self)


class BaseConfig(MutableMapping):
    def __init__(self, values=None, priority="project"):
        self.attributes = {}
        self.update(values, priority)

    def __getitem__(self, opt_name):
        if opt_name not in self:
            return None
        return self.attributes[opt_name].value

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def getbool(self, name, default=None):
        v = self.get(name, default)
        try:
            return bool(int(v))
        except ValueError:
            if v in ("True", "true"):
                return True
            if v in ("False", "false"):
                return False
        return None

    def getint(self, name, default=None):
        v = self.get(name, default)
        try:
            return int(v)
        except ValueError:
            pass
        return None

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        try:
            return float(v)
        except ValueError:
            pass
        return None

    def getlist(self, name, default=None):
        v = self.get(name, default or [])
        if isinstance(v, str):
            v = v.split(",")
        elif not hasattr(v, "__iter__"):
            v = [v]
        return list(v)

    def getpriority(self, name):
        if name not in self:
            return None
        return self.attributes[name].priority

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, priority="project"):
        priority = get_config_priority(priority)
        if name not in self:
            if isinstance(value, ConfigAttribute):
                self.attributes[name] = value
            else:
                self.attributes[name] = ConfigAttribute(value, priority)
        else:
            self.attributes[name].set(value, priority)

    def update(self, values, priority="project"):
        if values is not None:
            if isinstance(values, BaseConfig):
                for name in values:
                    self.set(name, values[name], values.getpriority(name))
            else:
                for name, value in values.items():
                    self.set(name, value, priority)

    def delete(self, name, priority="project"):
        priority = get_config_priority(priority)
        if priority >= self.getpriority(name):
            del self.attributes[name]

    def __delitem__(self, name):
        del self.attributes[name]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)


class Config(BaseConfig):
    def __init__(self, values=None, priority="project"):
        super().__init__()
        for key in dir(defaultconfig):
            if key.isupper():
                self.set(key.lower(), getattr(defaultconfig, key), "default")
        self.update(values, priority)



# coding=utf-8

import logging

from xpaw.http import HttpRequest

log = logging.getLogger(__name__)


class DedupeMiddleware:
    def handle_output(self, response, result):
        return self._handle_result(result)

    def handle_start_requests(self, result):
        return self._handle_result(result)

    def _handle_result(self, result):
        for r in result:
            if isinstance(r, HttpRequest):
                if not self._is_dupe(r):
                    yield r
                else:
                    log.debug("Find the request (method={}, url={}) is duplicated".format(r.method, r.url))
            else:
                yield r

    def _is_dupe(self, request):
        raise NotImplementedError


class LocalSetDedupeMiddleware(DedupeMiddleware):
    def __init__(self):
        self._url_set = set()

    def _is_dupe(self, request):
        s = request.method + " " + request.url
        if s not in self._url_set:
            self._url_set.add(s)
            return False
        return True



# coding=utf-8

import pytest
from aiohttp.helpers import parse_mimetype

from xpaw.selector import Selector


class TestSelector:
    def test_selector_list(self):
        html = """<li>a</li><div><ul><li>b</li><li>c</li></ul></div><ul><li>d</li></ul>"""
        s = Selector(html)
        assert s.xpath("//li/text()")[0].text == "a"
        assert s.xpath("//li/text()")[-1].text == "d"
        assert s.xpath("//div").xpath(".//li").text[0] == "b"
        assert s.xpath("//div").xpath(".//li").text[-1] == "c"

    def test_attribute_selection(self):
        html = """<div style=display:none><a href="http://example.com/" target=_blank></div>"""
        s = Selector(html)
        assert s.xpath("//a/@href")[0].html == "http://example.com/"
        assert s.xpath("//a").xpath("@target")[0].html == "_blank"
        assert s.xpath("//div/@style")[0].html == "display:none"

    def test_text_selection(self):
        html = """<div><p>expression: <var>x</var>+<var>y</var>=<var>z</var></p></div>"""
        s = Selector(html)
        assert s.xpath("//var[last()-1]/text()")[0].html == "y"
        assert s.xpath("//var/text()").html == ["x", "y", "z"]
        assert s.xpath("//p")[0].html == "<p>expression: <var>x</var>+<var>y</var>=<var>z</var></p>"
        assert s.xpath("//p")[0].text == "expression: x+y=z"
        assert s.xpath("//p/text()")[0].html == "expression: "

    def test_encoding_detection(self):
        html = "<html lang=en><head>" \
               "<title>测试</title>" \
               "<meta charset=utf-8>" \
               "<meta http-equiv=Content-Type content='text/html; charset=gbk' />" \
               "</head></html>"
        body = html.encode("gbk")
        with pytest.raises(UnicodeDecodeError):
            s = Selector(body.decode("utf-8"))
        s = Selector(body.decode("ascii", errors="ignore"))
        assert s.xpath("//meta/@charset")[0].html == "utf-8"
        content_type = s.xpath("//meta[@http-equiv='Content-Type']").xpath("@content")[0].html
        assert content_type == "text/html; charset=gbk"
        mtype, stype, _, params = parse_mimetype(content_type)
        assert mtype == "text" and stype == "html" and params.get("charset") == "gbk"

    def test_wrong_arguments(self):
        html = b"<html></html>"
        with pytest.raises(TypeError):
            s = Selector(html)
        with pytest.raises(ValueError):
            s = Selector()

    def test_confused_xpath(self):
        html = "<p>text</p><ul><p>header</p><li>a</li><li>b</li></ul>"
        s = Selector(html)

        assert s.xpath("/li") == []
        assert s.xpath("//li") != []

        assert s.xpath("/ul") == []
        assert s.xpath("/html/body/ul") != []

        assert s.xpath("//ul").xpath("//p")[0].text == "text"
        assert s.xpath("//ul").xpath(".//p")[0].text == "header"
        assert s.xpath("//ul").xpath("./p")[0].text == "header"



# coding=utf-8

import re
import logging

from xpaw.config import BaseConfig
from xpaw.errors import ResponseNotMatch, IgnoreRequest, NetworkError

log = logging.getLogger(__name__)


class RetryMiddleware:
    RETRY_ERRORS = (NetworkError, ResponseNotMatch)
    RETRY_HTTP_STATUS = (500, 502, 503, 504, 408)

    def __init__(self, max_retry_times, http_status):
        self._max_retry_times = max_retry_times
        self._http_status = http_status

    @classmethod
    def from_config(cls, config):
        c = config.get("retry")
        if c is None:
            c = {}
        c = BaseConfig(c)
        return cls(c.getint("max_retry_times", 3),
                   c.getlist("http_status", cls.RETRY_HTTP_STATUS))

    async def handle_response(self, request, response):
        for p in self._http_status:
            if self.match_status(p, response.status):
                return self.retry(request, "http status={}".format(response.status))

    @staticmethod
    def match_status(pattern, status):
        if isinstance(pattern, int):
            return pattern == status
        verse = False
        if pattern.startswith("!") or pattern.startswith("~"):
            verse = True
            pattern = pattern[1:]
        s = str(status)
        n = len(s)
        match = True
        if len(pattern) != n:
            match = False
        else:
            i = 0
            while i < n:
                if pattern[i] != "x" and pattern[i] != "X" and pattern[i] != s[i]:
                    match = False
                    break
                i += 1
        if verse:
            match = not match
        return match

    async def handle_error(self, request, error):
        if isinstance(error, self.RETRY_ERRORS):
            return self.retry(request, "{}: {}".format(type(error), error))

    def retry(self, request, reason):
        retry_times = request.meta.get("_retry_times", 0) + 1
        if retry_times <= self._max_retry_times:
            log.debug("We will retry the request(url={}) because of {}".format(request.url, reason))
            request.meta["_retry_times"] = retry_times
            return request.copy()
        else:
            log.info("The request(url={}) has been retried {} times,"
                     " and it will be aborted.".format(request.url, self._max_retry_times))
            raise IgnoreRequest()


class ResponseMatchMiddleware:
    def __init__(self, url_pattern, body_pattern, encoding):
        if not url_pattern:
            raise ValueError("url pattern is none")
        if not body_pattern:
            raise ValueError("body pattern is none")
        self._url_pattern = re.compile(url_pattern)
        self._body_pattern = re.compile(body_pattern)
        self._encoding = encoding

    @classmethod
    def from_config(cls, config):
        c = config.get("response_match")
        if c is None:
            c = {}
        c = BaseConfig(c)
        return cls(c.get("url_pattern"),
                   c.get("body_pattern"),
                   c.get("encoding"))

    async def handle_response(self, request, response):
        req_url = str(request.url)
        if response.body:
            if self._url_pattern.search(req_url):
                if self._encoding:
                    text = response.body.decode(self._encoding, errors="replace")
                else:
                    text = response.text
                if not self._body_pattern.search(text):
                    raise ResponseNotMatch("Response body does not fit the pattern")



# coding=utf-8

import time
import socket


def wait_server_start(addr):
    host, port = addr.split(":")
    port = int(port)
    not_ready = True
    while not_ready:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        res = s.connect_ex((host, port))
        s.close()
        if res == 0:
            not_ready = False
        else:
            time.sleep(0.1)



# coding=utf-8

import logging

from xpaw.middleware import MiddlewareManager

log = logging.getLogger(__name__)


class Spider:
    def __init__(self, config):
        self.config = config

    @property
    def logger(self):
        return log

    def log(self, message, level=logging.INFO, **kw):
        self.logger.log(level, message, **kw)

    def parse(self, response):
        raise NotImplementedError

    def start_requests(self):
        raise NotImplementedError


def _isiterable(obj):
    return hasattr(obj, "__iter__")


class SpiderMiddlewareManager(MiddlewareManager):
    def __init__(self, *middlewares):
        self._input_handlers = []
        self._output_handlers = []
        self._error_handlers = []
        self._start_requests_handlers = []
        super().__init__(*middlewares)

    def _add_middleware(self, middleware):
        super()._add_middleware(middleware)
        if hasattr(middleware, "handle_input"):
            self._input_handlers.append(middleware.handle_input)
        if hasattr(middleware, "handle_output"):
            self._output_handlers.append(middleware.handle_output)
        if hasattr(middleware, "handle_error"):
            self._error_handlers.append(middleware.handle_error)
        if hasattr(middleware, "handle_start_requests"):
            self._start_requests_handlers.append(middleware.handle_start_requests)

    @classmethod
    def _middleware_list_from_config(cls, config):
        mw_list = config.get("spider_middlewares")
        if mw_list:
            if not isinstance(mw_list, list):
                mw_list = [mw_list]
        else:
            mw_list = []
        log.debug("Spider middleware list: {}".format(mw_list))
        return mw_list

    def parse(self, spider, response):
        try:
            self._handle_input(response)
            if response.request and response.request.callback:
                r = getattr(spider, response.request.callback)(response)
            else:
                r = spider.parse(response)
            if r:
                r = self._handle_output(response, r)
            if r is None:
                return ()
            for i in r:
                yield i
        except Exception as e:
            try:
                handled = self._handle_error(response, e)
            except Exception:
                raise
            else:
                if handled is not True:
                    raise e

    def start_requests(self, spider):
        try:
            r = spider.start_requests()
            if r:
                r = self._handle_start_requests(r)
            if r is None:
                return ()
            for i in r:
                yield i
        except Exception as e:
            yield e

    def _handle_input(self, response):
        for method in self._input_handlers:
            res = method(response)
            if res is not None:
                raise TypeError("Input handler must return None, got {}".format(type(res)))

    def _handle_output(self, response, result):
        for method in self._output_handlers:
            result = method(response, result)
            if not _isiterable(result):
                raise TypeError("Response handler must return an iterable object, got {}".format(type(result)))
        return result

    def _handle_error(self, response, error):
        handled = False
        for method in self._error_handlers:
            res = method(response, error)
            if not (res is None or res is True):
                raise TypeError("Exception handler must return None or True, got {}".format(type(res)))
            if res is True:
                handled = True
        return handled

    def _handle_start_requests(self, result):
        for method in self._start_requests_handlers:
            result = method(result)
            if not _isiterable(result):
                raise TypeError("Start requests handler must return an iterable object, got {}".format(type(result)))
        return result



# coding=utf-8

from xpaw.config import Config
from xpaw.spidermws import *
from xpaw.http import HttpRequest, HttpResponse


class TestDepthMiddleware:
    max_depth = 1
    key = "_current_depth"

    def test_handle_output(self):
        class R:
            def __init__(self, depth=None):
                self.meta = {}
                if depth is not None:
                    self.meta["_current_depth"] = depth

        mw = MaxDepthMiddleware.from_config(Config({"max_depth": 1}))
        req = HttpRequest("http://127.0.0.1", "GET")
        resp = HttpResponse("http://127.0.0.1", 200)
        res = [i for i in mw.handle_output(R(), [req, resp])]
        assert res == [req, resp] and req.meta[self.key] == 1
        res = [i for i in mw.handle_output(R(0), [req, resp])]
        assert res == [req, resp] and req.meta[self.key] == 1
        res = [i for i in mw.handle_output(R(1), [req, resp])]
        assert res == [resp]
